/*
Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Amsterdam nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*! \file PlayerTeams.cpp
<pre>
<b>File:</b>          PlayerTest.cpp
<b>Project:</b>       Robocup Soccer Simulation Team: nopporoes4
<b>Authors:</b>       matu & yoko
<b>Created:</b>       1/30/2005
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This file contains the class definitions for the
                      Player that are used to test the teams' high level
                      strategy.
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
10/12/2000        Jelle Kok       Initial version created
</pre>
*/

#include "Player.h"

/*!This method is the first complete simple team and defines the actions taken
   by all the players on the field (excluding the goalie). It is based on the
   high-level actions taken by the simple team FC Portugal that it released in
   2000. The players do the following:
   - if ball is kickable
       kick ball to goal (random corner of goal)
   - else if i am fastest player to ball
       intercept the ball
   - else
       move to strategic position based on your home position and pos ball */


void Player::setFormationAfterKickOff() {
	
	PlayModeT playMode = WM->getPlayMode();
	double ballPosX = WM->getBallPos().getX();
	
	if ( WM->getSide() == SIDE_RIGHT ) {
		//ballPosX *= (-1.0);
		if ( ballPosX < 1.0 ) {
			formations->setFormation( FT_IRON_DEFENSIVE );     //( FT_334_OFENSIVE ) ¦λ //( FT_IRON_DEFENSIVE )¦λ
		}
		else {
			formations->setFormation( FT_343_ATTACKING );     //( FT_433_OFFENSIVE )¦λ//( FT_343_ATTACKING ) ¦λ
		}

	}
	else{

		if ( ballPosX < -1.0 ) {
			formations->setFormation( FT_IRON_DEFENSIVE );     //( FT_334_OFENSIVE ) ¦λ //( FT_IRON_DEFENSIVE )¦λ
		}
		else {
			formations->setFormation( FT_343_ATTACKING );     //( FT_433_OFFENSIVE )¦λ//( FT_343_ATTACKING ) ¦λ
		}
	}
}

/////////////////////////////////////////////////
/////////////////////////////////////////////////
//PLAYER  FW  FW_W_R  FW_W_L  MF  DF_SW  DF_C  DF_W
/////////////////////////////////////////////////
/////////////////////////////////////////////////
SoccerCommand Player::FW(  )
{

  SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  VecPosition   posEnd   = WM->getGlobalPosition( OBJECT_FLAG_P_R_C );
  int           iTmp;
  AngDeg        angMax;

  if( WM->isBeforeKickOff( ) )
  {
    if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    {
      if( WM->isBallKickable() )
      {
        
        soc = leadingPass( OBJECT_TEAMMATE_2, 0.0, DIR_NORTH ); // kick maximal
        Log.log( 100, "take kick off" );
      }
      else
      {
        soc = intercept( false );
        Log.log( 100, "move to ball to take kick-off" );
      }
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      return soc;
    }
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos( WM->getStrategicPosition() ));
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = searchBall() );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else
  {
    setFormationAfterKickOff();
    soc.commandType = CMD_ILLEGAL;

    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    }
    else if( WM->isBallKickable())                    // if kickable
    {

      printf( "ball %d\n", posAgent.getY() );
      VecPosition posGoal( PITCH_LENGTH / 2.0, SS->getGoalWidth() * 0.45 );
      if( posAgent.getY() < 0 )
      {
        VecPosition posGoal( PITCH_LENGTH/2.0, -SS->getGoalWidth() * 0.45 );
      }
 /*
  {
  VecPosition posShoot = getThroughPassShootingPoint( o, posEnd, angMax );
  double      dEnd     = getEndSpeedForPass( o, posShoot );

  return kickTo( posShoot, dEnd );
  }
  */
     if( WM->isInTheirPenaltyArea( posBall ) )
     {
        soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal

        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        Log.log( 100, "kick ball" );
     }
     /*else
     {
       VecPosition posTeam  = WM->getGlobalPosition( OBJECT_TEAMMATE_11 );
       VecPosition posEnd;
       VecPosition posShoot = getThroughPassShootingPoint( OBJECT_TEAMMATE_11,
                                                           VecPosition( 45, 0 ),
                                                           &angMax );
       double      dEnd     = getEndSpeedForPass( OBJECT_TEAMMATE_11, posShoot );

       soc = kickTo( posShoot, dEnd );
       ACT->putCommandInQueue( soc );
     }*/
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
             SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
     }
     else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
     {
       if( WM->getAgentStamina().getStamina() >     // if stamina high
                            SS->getRecoverDecThr()*SS->getStaminaMax()+800 )
       {
         soc = moveToPos(WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
         ACT->putCommandInQueue( soc );            // move to strategic pos
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else                                        // else watch ball
       {
         ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
     }
     else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     {
       ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     }
     else                                         // nothing to do
       ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
   }
  return soc;
}
  
///////////////////////////////////////////////////
///////////////player 6,9//////////////////////////
///////////////////////////////////////////////////

 
SoccerCommand Player::FW_W_L(  )
{
  SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  VecPosition   posEnd   = WM->getGlobalPosition( OBJECT_FLAG_P_R_C );
  VecPosition   posGoal( PITCH_LENGTH/2.0,SS->getGoalWidth()*0.45 );
  if(posAgent.getY()<0){
        VecPosition posGoal( PITCH_LENGTH/2.0,-SS->getGoalWidth()*0.45 );
  }
  int           iTmp;
  AngDeg        angMax;

  if( WM->isBeforeKickOff( ) )
  {
    if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    {
      if( WM->isBallKickable() )
      {
        soc = leadingPass( OBJECT_TEAMMATE_2,0.0,DIR_NORTH ); // kick maximal
        Log.log( 100, "take kick off" );
      }
      else
      {
        soc = intercept( false );
        Log.log( 100, "move to ball to take kick-off" );
      }
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      return soc;
    }
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos( WM->getStrategicPosition() ));
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = searchBall() );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else   //
  {
    setFormationAfterKickOff();
    soc.commandType = CMD_ILLEGAL;

    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    }
    else if( WM->isBallKickable())                    // if kickable
    {
      if( WM->isFreeKickUs() || WM->isCornerKickUs() || WM->isOffsideUs() ||
                  WM->isKickInUs() )
      {
        soc = directPass(WM->getGlobalPosition( OBJECT_FLAG_P_R_C ),PASS_FAST);
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        Log.log( 100, "take kick off" );

      }
      else if( WM->isInTheirPenaltyArea( posBall ) )
      {
      
          // ġѹ
     // if( WM->getCurrentCycle() % 2 == 1 )
      if( WM->getScoreToPassRoute( posGoal ) > 2.0 )
        {
          soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
        }
        /*else
        {
          // soc = directPass( WM->getGlobalPosition( OBJECT_FLAG_P_R_C), PASS_FAST );
          soc = kickTo( heuristicPassPosition2( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), 5.0 ),
                        PS->getPassEndSpeed() );
        }*/
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        Log.log( 100, "kick ball" );
       }
       else if( WM->getPlayerNumber() == 6 )
       {
//         if( WM->isEmptySpace( OBJECT_TEAMMATE_6,   4, 4.0 ) &&
//             WM->isEmptySpace( OBJECT_TEAMMATE_6, 180, 0.4 ) &&
//             WM->isEmptySpace( OBJECT_TEAMMATE_6,  90, 1.0 ) &&
//             WM->isEmptySpace( OBJECT_TEAMMATE_6, -90, 1.0 ) )
	  if(WM->isEmptySpace( OBJECT_TEAMMATE_10, 180, 1 ) &&
             WM->isEmptySpace( OBJECT_TEAMMATE_10, -90, 1 ) &&
	     posAgent.getX() > 10)
	  {
		soc = leadingPass( OBJECT_TEAMMATE_10,0.0,DIR_NORTH );
	  }
	  else if(WM->isEmptySpace( OBJECT_TEAMMATE_11, 180, 1 ) &&
                    WM->isEmptySpace( OBJECT_TEAMMATE_11,  90, 1 ) &&
	            posAgent.getX() > 10)
	  {
		soc = leadingPass( OBJECT_TEAMMATE_11,0.0,DIR_NORTH );
	  }
	  else if( WM->isEmptySpace( OBJECT_TEAMMATE_6,   0, 4.0 ) &&
		    WM->isEmptySpace( OBJECT_TEAMMATE_6, 60,4,1))
         {
//           soc = dribble( -20, DRIBBLE_SLOW );
	     soc = dribble(-45,DRIBBLE_FAST);
           if( posAgent.getX() > 5*PITCH_LENGTH / 12.0 )
           {
 soc = directPass(heuristicPassPosition2( OBJECT_TEAMMATE_10, 5.0 ),PASS_FAST);
             //soc = kickTo( heuristicPassPosition2( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), 5.0 ),
             //              PS->getPassEndSpeed() );
           }
           else if(posAgent.getX() > 3 * PITCH_LENGTH / 10.0 )
           {
             soc = dribble( 15,DRIBBLE_WITHBALL );
           }
           ACT->putCommandInQueue( soc );
         }
//         else if(WM->isEmptySpace( OBJECT_TEAMMATE_6,2,2) &&
//                                   WM->isEmptySpace( OBJECT_TEAMMATE_6, 180, 1.0 ) &&
//                                   WM->isEmptySpace( OBJECT_TEAMMATE_6,  90, 1.0 ) &&
//                                   WM->isEmptySpace( OBJECT_TEAMMATE_6, -90, 1.0 ) )
	 else if(WM->isEmptySpace( OBJECT_TEAMMATE_6,   0, 4.0 ) &&
	  WM->isEmptySpace( OBJECT_TEAMMATE_6,0,4,1))
         {
//           soc = dribble( -14, DRIBBLE_SLOW );
	    soc = dribble(45,DRIBBLE_SLOW);
           if(posAgent.getX() > 5*PITCH_LENGTH/12.0)
           {
             // soc = directPass(WM->getGlobalPosition(OBJECT_FLAG_P_R_C),PASS_FAST);
             soc = kickTo( heuristicPassPosition2( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), 5.0 ),
                           PS->getPassEndSpeed() );
           }
           else if( posAgent.getX() > 3 * PITCH_LENGTH / 10.0 )
           {
             soc = dribble( 15, DRIBBLE_WITHBALL );
           }
	   if( WM->getScoreToPassRoute( posGoal ) > 2.0 )
           {
              soc = kickTo( posGoal, SS->getBallSpeedMax() );
           }
           ACT->putCommandInQueue( soc );
         }
	 else if(WM->isEmptySpace( OBJECT_TEAMMATE_6,0,2,1))
         {
           //soc = dribble( 14, DRIBBLE_SLOW );
	   soc = dribble(-45,DRIBBLE_WITHBALL);

           if( posAgent.getX() > 5 * PITCH_LENGTH / 12.0 )
           {
             // soc = directPass( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), PASS_FAST );
 soc = directPass(heuristicPassPosition2( OBJECT_TEAMMATE_10, 5.0 ),PASS_FAST);
//            soc = kickTo( heuristicPassPosition2( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), 5.0 ),
//                          PS->getPassEndSpeed() );
           }
           else if( posAgent.getX() > 3 * PITCH_LENGTH / 10.0 )
           {
             soc = dribble( -15, DRIBBLE_WITHBALL );
           }
           ACT->putCommandInQueue( soc );
         }
         else
         {
           // VecPosition posTeam10 = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );
           // soc = throughPass( OBJECT_TEAMMATE_10,posTeam10, NULL );
           // printf( "through pass!!! no.%d", WM->getPlayerNumber() );
           
             VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );
           if ( WM->getScoreToPassRoute( posPass ) > 7.0 ) 
             {

		soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_FAST );

             /*soc = kickTo( heuristicPassPosition2( posPass, 5.0 ),
                         getEndSpeedForPass( OBJECT_TEAMMATE_10, posPass ));*/
             }
           ACT->putCommandInQueue( soc );
         }
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_TEAMMATE_10, soc ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else if( WM->getPlayerNumber() == 9 )
       {
         if(WM->isEmptySpace( OBJECT_TEAMMATE_10, 180, 1 ) &&
             WM->isEmptySpace( OBJECT_TEAMMATE_10, -90, 1 ) &&
	     posAgent.getX() > 10)
	  {
VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );

		soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_FAST );
//		soc = leadingPass( OBJECT_TEAMMATE_10,0.0,DIR_NORTH );
	  }
	  else if(WM->isEmptySpace( OBJECT_TEAMMATE_11, 180, 1 ) &&
                    WM->isEmptySpace( OBJECT_TEAMMATE_11,  90, 1 ) &&
	            posAgent.getX() > 10)
	  {
VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_11 );

		soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_FAST );
//		soc = leadingPass( OBJECT_TEAMMATE_11,0.0,DIR_NORTH );
	  }
	  /*else if( WM->isEmptySpace( OBJECT_TEAMMATE_9,   0, 4.0 ) &&
		    WM->isEmptySpace( OBJECT_TEAMMATE_9, 60,4,1))
         {
//           soc = dribble( -20, DRIBBLE_SLOW );
	  //   soc = dribble(-45,DRIBBLE_FAST);
    /*       if( posAgent.getX() > 5*PITCH_LENGTH / 12.0 )
           {
             // soc = directPass(WM->getGlobalPosition(OBJECT_FLAG_P_R_C),PASS_FAST);
             soc = kickTo( heuristicPassPosition2( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), 5.0 ),
                           PS->getPassEndSpeed() );
           }
           else if(posAgent.getX() > 3 * PITCH_LENGTH / 10.0 )
           {
             soc = dribble( 15,DRIBBLE_WITHBALL );
           }*/
         //  ACT->putCommandInQueue( soc );
         //}
//         else if( WM->isEmptySpace( OBJECT_TEAMMATE_9,   2, 2)   &&
//                  WM->isEmptySpace( OBJECT_TEAMMATE_9, 180, 1.0) &&
//                  WM->isEmptySpace( OBJECT_TEAMMATE_9,  90, 1.0) &&
//                  WM->isEmptySpace( OBJECT_TEAMMATE_9, -90, 1.0) )
	  /*else if(WM->isEmptySpace( OBJECT_TEAMMATE_9,   0, 4.0 ) &&
	  WM->isEmptySpace( OBJECT_TEAMMATE_9,0,4,1))
         {
//           soc = dribble( -14, DRIBBLE_SLOW );
	    soc = dribble(45,DRIBBLE_SLOW);
           if(posAgent.getX() > 5*PITCH_LENGTH/12.0)
           {
             // soc = directPass(WM->getGlobalPosition(OBJECT_FLAG_P_R_C),PASS_FAST);
VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );
 soc = directPass(heuristicPassPosition2( posPass, 5.0 ),PASS_FAST);
//             soc = kickTo( heuristicPassPosition2( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), 5.0 ),
//                           PS->getPassEndSpeed() );
           }
          /* else if( posAgent.getX() > 3 * PITCH_LENGTH / 10.0 )
           {
             soc = dribble( 15, DRIBBLE_WITHBALL );
           }
	   if( WM->getScoreToPassRoute( posGoal ) > 2.0 )
           {
              soc = kickTo( posGoal, SS->getBallSpeedMax() );
           }
           ACT->putCommandInQueue( soc );
         }
/*	 else if(WM->isEmptySpace( OBJECT_TEAMMATE_9,0,2,1))
         {
           //soc = dribble( -14, DRIBBLE_SLOW );
	   soc = dribble(-45,DRIBBLE_WITHBALL);
           if(posAgent.getX() > 5*PITCH_LENGTH/12.0)
           {
             // soc = directPass(WM->getGlobalPosition(OBJECT_FLAG_P_R_C),PASS_FAST);
             soc = kickTo( heuristicPassPosition2( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), 5.0 ),
                           PS->getPassEndSpeed() );
           }
           else if( posAgent.getX() > 3 * PITCH_LENGTH / 10.0 )
           {
             soc = dribble( 15, DRIBBLE_WITHBALL );
           }
           ACT->putCommandInQueue( soc );
         }*/
         else
         {
           // VecPosition posTeam11 = WM->getGlobalPosition( OBJECT_TEAMMATE_11 );
           // soc = throughPass( OBJECT_TEAMMATE_11,posTeam11, NULL );
           VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );

	    soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_FAST );

          /* soc = kickTo( heuristicPassPosition2( posPass, 5.0 ),
                         getEndSpeedForPass( OBJECT_TEAMMATE_10, posPass ));*/
           printf("through pass!!! no.%d",WM->getPlayerNumber());
           ACT->putCommandInQueue( soc );
         }
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_TEAMMATE_11, soc ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
          SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
             1.5 + fabs( posAgent.getX() - posBall.getX() ) / 10.0 )
                                                  // if not near strategic pos
     {
       if( WM->getAgentStamina().getStamina() >     // if stamina high
           SS->getRecoverDecThr()*SS->getStaminaMax()+800 )
       {
         soc = moveToPos(WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
         ACT->putCommandInQueue( soc );            // move to strategic pos
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else                                        // else watch ball
       {
         ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
     }
     else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     {
       ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     }
     else {                                    // nothing to do
       ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
     }
  }
  return soc;
}
/////////////////////////////////////////////////
//////////////player 10,11///////////////////
/////////////////////////////////////////////////
SoccerCommand Player::FW_W_R(  )
{

  SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  VecPosition   posGoal( PITCH_LENGTH/2.0,SS->getGoalWidth()*0.45 );
  if( posAgent.getY() < 0 )
  {
     VecPosition posGoal( PITCH_LENGTH / 2.0, -SS->getGoalWidth() * 0.45 );
  }
  int           iTmp,hi=0,himax=0,i;


  if( WM->isBeforeKickOff( ) )
  {
    if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    {
      if( WM->isBallKickable() ) {
      /*
        VecPosition posGoal( PITCH_LENGTH / 2.0,
                              ( -1 + 2 * ( WM->getCurrentCycle() % 2 ) ) *
                               0.4 * SS->getGoalWidth() );
        soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
      */
        soc = leadingPass( OBJECT_TEAMMATE_2, 0.0, DIR_NORTH );
        Log.log( 100, "take kick off" );
      }
      else
      {
        soc = intercept( false );
        Log.log( 100, "move to ball to take kick-off" );
      }
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      return soc;
    }
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos( WM->getStrategicPosition() ));
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = searchBall() );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else
  { 
    setFormationAfterKickOff();
    soc.commandType = CMD_ILLEGAL;

    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    }
    else if( WM->isBallKickable())                    // if kickable
    {
      if( WM->isFreeKickUs() || WM->isCornerKickUs() || WM->isOffsideUs() ||
                  WM->isKickInUs() )
      {
        // soc = directPass( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), PASS_FAST );
          // Ϥ˲̿᤬ɬ
        soc = kickTo( heuristicPassPosition2( WM->getGlobalPosition( OBJECT_FLAG_P_R_C ), 5.0 ),
                      PS->getPassEndSpeed() );
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        Log.log( 100, "take kick off" );

      }
     else
      {
      if( posBall.getX() >= ( PITCH_LENGTH / 2 - PENALTY_AREA_LENGTH ) )
        {
        if( WM->isInTheirPenaltyArea( posBall ) )
          {
          if( WM->getPlayerNumber() == 11 )
            {
            if( WM->isEmptySpace( OBJECT_TEAMMATE_11, 0, 2 ) && posBall.getX() <= 45)
              {
               // if( WM->getCurrentCycle() % 2 == 1 )
//             if( WM->getScoreToPassRoute( posGoal ) > 2.0 )
//                {
                  soc = kickTo( posGoal, SS->getBallSpeedMax() );
//                }
//                else {
                  // soc = leadingPass( OBJECT_TEAMMATE_10, 0.7 ,DIR_NORTH );
//                  VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );

//		soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );

                 /* soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                                getEndSpeedForPass( OBJECT_TEAMMATE_10, posPass ));*/
//                }
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );

              }
              else if( posBall.getY() >= - ( PENALTY_AREA_WIDTH / 4.0 + 5 ) )
              {
//	        if( WM->getScoreToPassRoute( posGoal ) > 1.0 )
//                {
                  soc = kickTo( posGoal, SS->getBallSpeedMax() );
//                }
             // if( WM->getCurrentCycle() % 2 == 1 )
//             if( WM->getScoreToPassRoute( posGoal ) > 2.0 )
//                {
//                  soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
//                }
//             else
//                {
                  // soc = leadingPass( OBJECT_TEAMMATE_10, 0.7 ,DIR_NORTH);
//                  VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );

//		soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );

                  /*soc = kickTo( heuristicPassPosition2( posPass, 5.0 ),
                                getEndSpeedForPass( OBJECT_TEAMMATE_10, posPass ) );*/
//                }
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
              }
              else if( WM->isEmptySpace( OBJECT_TEAMMATE_10, -90, 2 ) )
              {
                // soc = leadingPass( OBJECT_TEAMMATE_10, 0.7 ,DIR_NORTH);
                VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );

		soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );

                /*soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                              getEndSpeedForPass( OBJECT_TEAMMATE_10, posPass ));*/

                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_TEAMMATE_10, soc ) );
              }
              else if( posAgent.getDistanceTo(
                           WM->getGlobalPosition( OBJECT_FLAG_G_R_T ) ) <= 13 )
              {
                for( i  = 1;
                     i <= posAgent.getDistanceTo( WM->getGlobalPosition( OBJECT_FLAG_G_R_T ) );
                     i++)
                {
                  if( WM->isEmptySpace( OBJECT_TEAMMATE_11,
                                        WM->getGlobalPosition( OBJECT_FLAG_G_R_T ).getDirection(),
                                        i ) )
                  {
                    hi++;
                  }
                  himax++;
                }
                if( hi == himax )
                {
                  ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
                  VecPosition posGoal( PITCH_LENGTH / 2.0,
                                       ( -1 + 2 * ( WM->getCurrentCycle() % 2 ) *posAgent.getX()) *
                                       0.4 * SS->getGoalWidth() );
                  soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
                  ACT->putCommandInQueue( soc );
                }
              }
            }
            else //player_10
            {
              // ʬ-30٤2mΥڡä43m
              if( WM->isEmptySpace( OBJECT_TEAMMATE_10, -30, 2 ) &&
                  posBall.getX() <= 43 )
              {
//                if( WM->getCurrentCycle() % 2 == 1 )
//                {
                  soc = kickTo( posGoal, SS->getBallSpeedMax());
//                }
//                else
//                {
                  // soc = leadingPass( OBJECT_TEAMMATE_10, 0.7 ,DIR_NORTH);
//                  VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_11 );

//		  soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );

                  /*soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                                getEndSpeedForPass( OBJECT_TEAMMATE_10, posPass ));*/
//                }
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
              }
              else if( posBall.getY() <= ( PENALTY_AREA_WIDTH / 4.0 + 5 ) )
              {
                // ġѹ
             // if( WM->getCurrentCycle() % 3 == 1 )
//             if( WM->getScoreToPassRoute( posGoal ) > 1.0 )
//                {
                  soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
//                }
//             else 
//                {
                  // soc = leadingPass( OBJECT_TEAMMATE_11, 1.1, DIR_NORTH );
//                  VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_11 );

//		soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );

                 /* soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                                getEndSpeedForPass( OBJECT_TEAMMATE_11, posPass ));*/
               
//                }
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
              }
              else if( WM->isEmptySpace( OBJECT_TEAMMATE_11, 90, 2 ) )
              {
                // soc = leadingPass( OBJECT_TEAMMATE_11, 1.1, DIR_NORTH );
                VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_11 );

		soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );

                /*soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                              getEndSpeedForPass( OBJECT_TEAMMATE_11, posPass ));*/
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_TEAMMATE_11, soc ) );
              }
              else
              {
                soc = holdBall();
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
              }
              if( posAgent.getDistanceTo( WM->getGlobalPosition( OBJECT_FLAG_G_R_B ) ) <= 13 )
              {
                for( i  = 1;
                     i <= posAgent.getDistanceTo( WM->getGlobalPosition( OBJECT_FLAG_G_R_B ) );
                     i++)
                {
                  if( WM->isEmptySpace( OBJECT_TEAMMATE_10,
                                        WM->getGlobalPosition( OBJECT_FLAG_G_R_B ).getDirection(),
                                        i ) )
                  {
                    hi++;
                  }
                  himax++;
                }
                if( hi == himax )
                {
                  ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
                  VecPosition posGoal( PITCH_LENGTH / 2.0,
                                       ( -1 + 2 * ( WM->getCurrentCycle() % 2 ) ) * 0.4 * SS->getGoalWidth() );
                  soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
                  ACT->putCommandInQueue( soc );
                }
              }
            }
          }
          else
          {
            if(posBall.getY() < 0)
            {
              soc = dribble(  60, DRIBBLE_SLOW );
            }
            else
            {
              soc = dribble( -57, DRIBBLE_SLOW );
            }
            ACT->putCommandInQueue( soc );
            ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );

          }
        }
        else
        {
          if(WM->getPlayerNumber() == 11)
          {
            if(WM->isEmptySpace( OBJECT_TEAMMATE_11,
                      ( -1 + 2 * ( WM->getCurrentCycle() % 2 ) ) * 0.4 * SS->getGoalWidth(),
                      2 ) )
            {

      //      soc = freezeBall( );
      //      ACT->putCommandInQueue( soc );
              // soc = leadingPass( OBJECT_TEAMMATE_9 ,2.3 ,DIR_NORTH);
               /* VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_9 );
                soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                              getEndSpeedForPass( OBJECT_TEAMMATE_9, posPass ));*/
              if( WM->isEmptySpace( OBJECT_TEAMMATE_9, 5, 4.0 ) )
              {

//if( posBall.getX() > 2 * PITCH_LENGTH / 10.0 )
/*if ( WM->getSide() == SIDE_LEFT && posBall.getX() > 2 * PITCH_LENGTH / 10.0 ) {
                

soc = dribble(20,DRIBBLE_SLOW);
//                  soc = kickTo( posGoal, SS->getBallSpeedMax() );
                
}else{



*/

                 soc = dribble( ( -1 + 2 * ( WM->getCurrentCycle() % 2 ) ) * 0.4 * SS->getGoalWidth(),
                                DRIBBLE_SLOW);

//}

              }
              ACT->putCommandInQueue( soc );
              ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
            }
            else
            {
              if(WM->isEmptySpace( OBJECT_TEAMMATE_10,
                                   ( -1 + 2 * ( WM->getCurrentCycle() % 2 ) ) * 0.4 * SS->getGoalWidth(),
                                   2 ) )
              {

                //soc = leadingPass( OBJECT_TEAMMATE_10 ,0.3 ,DIR_NORTH);

if( posBall.getX() > 2 * PITCH_LENGTH / 10.0 && WM->getScoreToPassRoute( posGoal ) > 1.0)
                {
                  soc = kickTo( posGoal, SS->getBallSpeedMax() );
                }

else{

                VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );

soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );
//soc = leadingPass( OBJECT_TEAMMATE_10 ,0.3 ,DIR_NORTH);

}
                /*soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                              getEndSpeedForPass( OBJECT_TEAMMATE_10, posPass ));*/
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_TEAMMATE_10, soc ) );
              }
           else if( WM->isEmptySpace( OBJECT_TEAMMATE_9,
                    ( -1 + 2 * (WM->getCurrentCycle() % 2 ) ) * 0.4 * SS->getGoalWidth(),
                                         2))
              {
                
                // soc = leadingPass( OBJECT_TEAMMATE_9 ,2.3 ,DIR_NORTH);
                if ( WM->getScoreToPassRoute( WM->getGlobalPosition( OBJECT_TEAMMATE_9 ) ) > 7.0 )
                    {
                      soc = directPass( OBJECT_TEAMMATE_9, PASS_NORMAL );
                    } 
                    /*else
                    {
                   
                  VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_9 );
                  soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                                getEndSpeedForPass( OBJECT_TEAMMATE_9, posPass ));
                     }*/
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_TEAMMATE_9, soc ) );
                  
              }
              else
              {
                soc = holdBall();
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
              }
            }
          }
          else
          {  //number10
            if(WM->isEmptySpace( OBJECT_TEAMMATE_10,
                      (-1 + 2*(WM->getCurrentCycle()%2)) * 0.4 * SS->getGoalWidth(),3))
            {

      //      soc = freezeBall( );
      //      ACT->putCommandInQueue( soc );
              // soc = leadingPass( OBJECT_TEAMMATE_6 ,2.3 ,DIR_NORTH);

              /*  VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_6 );
                soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                              getEndSpeedForPass( OBJECT_TEAMMATE_6, posPass ));*/
              if(WM->isEmptySpace( OBJECT_TEAMMATE_6,-5,2)){

//if( posBall.getX() > 2 * PITCH_LENGTH / 10.0 )
/*if ( WM->getSide() == SIDE_LEFT && posBall.getX() > 2 * PITCH_LENGTH / 10.0 ) {
                
soc = dribble(-20,DRIBBLE_SLOW);
                  //soc = kickTo( OBJECT_GOAL_R, SS->getBallSpeedMax() );
                }
else{


*/



                 soc = dribble( ( -1 + 2 * ( WM->getCurrentCycle() % 2 ) ) * 0.4 * SS->getGoalWidth(),
                                DRIBBLE_SLOW);

//}


              }
              ACT->putCommandInQueue( soc );              
              ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
            }
            else
            {
              if(WM->isEmptySpace( OBJECT_TEAMMATE_11,
                      (-1 + 2*(WM->getCurrentCycle()%2)) * 0.4 * SS->getGoalWidth(),1))
              {

if( posBall.getX() > 2 * PITCH_LENGTH / 10.0 && WM->getScoreToPassRoute( posGoal ) > 1.0)
                {
                  soc = kickTo( posGoal, SS->getBallSpeedMax() );
                }

else{               
                // soc = leadingPass( OBJECT_TEAMMATE_11 ,0.3 ,DIR_NORTH);
                VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_11 );

soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );
// soc = leadingPass( OBJECT_TEAMMATE_11 ,0.3 ,DIR_NORTH);

}
                /*soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                              getEndSpeedForPass( OBJECT_TEAMMATE_11, posPass ));*/
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_TEAMMATE_11, soc ) );
              }
              /*else if(WM->isEmptySpace( OBJECT_TEAMMATE_6,
                      (-1 + 2*(WM->getCurrentCycle()%2)) * 0.4 * SS->getGoalWidth(),2))
              {

                // soc = leadingPass( OBJECT_TEAMMATE_6 ,2.3 ,DIR_WEST);
                VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_6 );
                soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                              getEndSpeedForPass( OBJECT_TEAMMATE_6, posPass ));
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_TEAMMATE_6, soc ) );
              }*/
              else
              {
                soc = holdBall();
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
              }
            }
          }
        }
      }  
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
             SS->getRecoverDecThr() * SS->getStaminaMax() + 100 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }

    }
    else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
    {
      if( WM->getAgentStamina().getStamina() >     // if stamina high
                  SS->getRecoverDecThr()*SS->getStaminaMax() + 700 )
      {
        soc = moveToPos( WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
        ACT->putCommandInQueue( soc );            // move to strategic pos
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                        // else watch ball
      {
        ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
    {
      ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else                                         // nothing to do
      ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
    }
    return soc;
}

/////////////////////////////////////////////////
/////////////////////////////////////////////////

SoccerCommand Player::MF(  )
{
 SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  int           iTmp;


  VecPosition   posGoal( PITCH_LENGTH/2.0,SS->getGoalWidth()*0.45 );
  if( posAgent.getY() < 0 )
  {
     VecPosition posGoal( PITCH_LENGTH / 2.0, -SS->getGoalWidth() * 0.45 );
  }

  if( WM->isBeforeKickOff( ) )
  {
    if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    {
      if( WM->isBallKickable() )
      {
        cerr << "test" << endl;
        soc = leadingPass( OBJECT_TEAMMATE_2, 0.0, DIR_NORTH ); // kick maximal
        Log.log( 100, "take kick off" );
        
      }
      else
      {
        soc = intercept( false );
        Log.log( 100, "move to ball to take kick-off" );
      }
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      return soc;
    }
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc = teleportToPos( WM->getStrategicPosition() ));
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = searchBall() );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else
  {
      setFormationAfterKickOff();
      
    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    }
    else if( WM->isBallKickable())                    // if kickable
    {
        // ġɲ
      if ( posAgent.getX() > 38.0 ) {
        if( WM->getScoreToPassRoute( posGoal ) > 2.0 )
          {
          soc = kickTo( posGoal, SS->getBallSpeedMax() );
          }
        }
        else 
        {
    
      // VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_6 );

if( posBall.getX() > 4 * PITCH_LENGTH / 10.0 )
                {
                  
 if(WM->getPlayerNumber() == 7)
      {
VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_10 );
soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );
//          soc = HeuristicPass2( WM->getGlobalPosition( OBJECT_TEAMMATE_6 ),
//                                3.0 );
      }
      else if(WM->getPlayerNumber() == 8)
      {
VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_11 );
soc = directPass( heuristicPassPosition2( posPass, 5.0 ),PASS_NORMAL );
//          soc = HeuristicPass2( WM->getGlobalPosition( OBJECT_TEAMMATE_10 ),
//                               3.0 );
      }



                }
else{

soc = directPass( heuristicPassPosition( ),PASS_NORMAL );
//        soc = kickTo( heuristicPassPosition( ), PS->getPassEndSpeed() );
}
        }
      /*
      if(WM->getPlayerNumber() == 7)
      {
          //soc = leadingPass( OBJECT_TEAMMATE_6 ,1.0 ,DIR_NORTH);
          soc = HeuristicPass2( WM->getGlobalPosition( OBJECT_TEAMMATE_6 ),
                                3.0 );
      }
      else if(WM->getPlayerNumber() == 8)
      {
          // soc = leadingPass( OBJECT_TEAMMATE_9 ,1.0 ,DIR_NORTH);
          soc = HeuristicPass2( WM->getGlobalPosition( OBJECT_TEAMMATE_10 ),
                                3.0 );
      }
      */
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      Log.log( 100, "kick ball" );
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
             SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
     }
     else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
     {
       if( WM->getAgentStamina().getStamina() >     // if stamina high
                            SS->getRecoverDecThr()*SS->getStaminaMax()+400 )
       {
         soc = moveToPos(WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
         ACT->putCommandInQueue( soc );            // move to strategic pos
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else                                        // else watch ball
       {
         ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
     }
     else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     {
       ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     }
     else                                         // nothing to do
       ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
   }
  return soc;
}

/////////////////////////////////////////////////
/////////////////////////////////////////////////

SoccerCommand Player::DF_CENTER( )
{

  SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  int           iTmp;

   if( WM->isBeforeKickOff( ) )
  {
    if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    {
      if( WM->isBallKickable() )
      {
        VecPosition posGoal( PITCH_LENGTH/2.0,
                             (-1 + 2*(WM->getCurrentCycle()%2)) *
                             0.4 * SS->getGoalWidth() );
        soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
        Log.log( 100, "take kick off" );
      }
      else
      {
        soc = intercept( false );
        Log.log( 100, "move to ball to take kick-off" );
      }
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      return soc;
    }
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos( WM->getStrategicPosition() ));
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = searchBall() );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else
  {
    setFormationAfterKickOff();
    soc.commandType = CMD_ILLEGAL;

    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    }
    else if( WM->isBallKickable())                    // if kickable
    {
//      soc = freezeBall( );
//      ACT->putCommandInQueue( soc );
      if( WM->isFreeKickUs() || WM->isCornerKickUs() || WM->isOffsideUs() ||
                  WM->isKickInUs() )
      {
        if(posBall.getY() < 0){
          // soc = leadingPass( OBJECT_TEAMMATE_6 ,2.1 ,DIR_NORTH);


          VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_6 );
      //    soc = kickTo( heuristicPassPosition2( posPass, 5.0),
      //                  getEndSpeedForPass( OBJECT_TEAMMATE_6, posPass ));
          ACT->putCommandInQueue( soc );
          ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        }
        else
        {
          // soc = leadingPass( OBJECT_TEAMMATE_7 ,2.1 ,DIR_NORTH);

          VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_7 );
          soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                        getEndSpeedForPass( OBJECT_TEAMMATE_7, posPass ));
          ACT->putCommandInQueue( soc );
          ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );        
        }
      }
      else
      {
        VecPosition posGoal( PITCH_LENGTH/2.0,
                (-1 + 2*(WM->getCurrentCycle()%2)) * 0.4 * SS->getGoalWidth() );
        soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
  
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        Log.log( 100, "kick ball" );
      }
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
             SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
     }
     else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
     {
       if( WM->getAgentStamina().getStamina() >     // if stamina high
                            SS->getRecoverDecThr()*SS->getStaminaMax()+800 )
       {
         soc = moveToPos(WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
         ACT->putCommandInQueue( soc );            // move to strategic pos
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else                                        // else watch ball
       {
         ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
     }
     else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     {
       ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     }
     else                                         // nothing to do
       ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
   }
  return soc;
}

/////////////////////////////////////////////////
/////////////////////////////////////////////////

SoccerCommand Player::DF_SWEEPER( )
{
   SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  int           iTmp;

  if( WM->isBeforeKickOff( ) )
  {
    if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    {
      if( WM->isBallKickable() )
      {
        VecPosition posGoal( PITCH_LENGTH/2.0,
                             (-1 + 2*(WM->getCurrentCycle()%2)) *
                             0.4 * SS->getGoalWidth() );
        soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
        Log.log( 100, "take kick off" );
      }
      else
      {
        soc = intercept( false );
        Log.log( 100, "move to ball to take kick-off" );
      }
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      return soc;
    }
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos( WM->getStrategicPosition() ));
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = searchBall() );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else
  {
    setFormationAfterKickOff();
    soc.commandType = CMD_ILLEGAL;

    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    }
    else if( WM->isBallKickable())                    // if kickable
    {
//      soc = freezeBall( );
//      ACT->putCommandInQueue( soc );
      if( WM->isFreeKickUs() || WM->isCornerKickUs() || WM->isOffsideUs() ||
                  WM->isKickInUs() )
      {
        if(posBall.getY() < 0){
          // soc = leadingPass( OBJECT_TEAMMATE_6 ,2.1 ,DIR_NORTH);


          VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_6 );
      //    soc = kickTo( heuristicPassPosition2( posPass, 5.0),
      //                  getEndSpeedForPass( OBJECT_TEAMMATE_6, posPass ));
          ACT->putCommandInQueue( soc );
          ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        }
        else
        {
          // soc = leadingPass( OBJECT_TEAMMATE_7 ,2.1 ,DIR_NORTH);

          VecPosition posPass = WM->getGlobalPosition( OBJECT_TEAMMATE_7 );
          soc = kickTo( heuristicPassPosition2( posPass, 5.0),
                        getEndSpeedForPass( OBJECT_TEAMMATE_7, posPass ));
          ACT->putCommandInQueue( soc );
          ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );        
        }
      }
      else
      {
        VecPosition posGoal( PITCH_LENGTH/2.0,
                (-1 + 2*(WM->getCurrentCycle()%2)) * 0.4 * SS->getGoalWidth() );
        soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
  
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        Log.log( 100, "kick ball" );
      }
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
             SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
     }
     else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
     {
       if( WM->getAgentStamina().getStamina() >     // if stamina high
                            SS->getRecoverDecThr()*SS->getStaminaMax()+800 )
       {
         soc = moveToPos(WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
         ACT->putCommandInQueue( soc );            // move to strategic pos
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else                                        // else watch ball
       {
         ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
     }
     else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     {
       ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     }
     else                                         // nothing to do
       ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
   }
  return soc;
}

/////////////////////////////////////////////////
/////////////////////////////////////////////////

SoccerCommand Player::DF_WING( )
{

  SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  int           iTmp,i=0;
  
  if( WM->isBeforeKickOff( ) )
  {
    if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    {
      if( WM->isBallKickable() )
      {
        VecPosition posGoal( PITCH_LENGTH/2.0,
                             (-1 + 2*(WM->getCurrentCycle()%2)) *
                             0.4 * SS->getGoalWidth() );
        soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
        Log.log( 100, "take kick off" );
      }
      else
      {
        soc = intercept( false );
        Log.log( 100, "move to ball to take kick-off" );
      }
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      return soc;
    }
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos( WM->getStrategicPosition() ));
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = searchBall() );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else
  {
    setFormationAfterKickOff();
    soc.commandType = CMD_ILLEGAL;
    ACT->putCommandInQueue( soc );
    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    } 
    else if(posBall.getX() < -(PITCH_LENGTH/2.0 + PENALTY_AREA_LENGTH + 5.5))
    {
      if( WM->getClosestRelativeInSet( OBJECT_SET_OPPONENTS , NULL ) == OBJECT_OPPONENT_11 )
      {
        soc = mark(OBJECT_OPPONENT_11, 0.4, MARK_GOAL);
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else if( WM->getClosestRelativeInSet( OBJECT_SET_OPPONENTS , NULL ) ==
               OBJECT_OPPONENT_10 )
      {
        soc = mark(OBJECT_OPPONENT_10, 0.5, MARK_GOAL);
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else if( WM->getClosestRelativeInSet( OBJECT_SET_OPPONENTS , NULL ) ==
               OBJECT_OPPONENT_9)
      {
        soc = mark( OBJECT_OPPONENT_9, 0.4, MARK_GOAL );
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( WM->isBallKickable() )                    // if kickable  2004/10/20
    {
      if( WM->isFreeKickUs() || WM->isCornerKickUs() || WM->isOffsideUs() ||
                  WM->isKickInUs() )
      {
          soc = clearBall( CLEAR_BALL_OFFENSIVE, SIDE_ILLEGAL, NULL );
        // soc = kickTo( heuristicPassPosition( ), PS->getPassEndSpeed() );
        /*
        if(WM->getPlayerNumber()==5)
          soc = leadingPass( OBJECT_TEAMMATE_9,0.1 ,DIR_NORTH);
        else if(WM->getPlayerNumber()==2)
          soc = leadingPass( OBJECT_TEAMMATE_7,0.1 ,DIR_NORTH);
        */
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        Log.log( 100, "take kick off" );

      }
      else
      {
        soc = clearBall(CLEAR_BALL_OFFENSIVE, SIDE_ILLEGAL, NULL );
//        soc = kickTo( heuristicPassPosition( ), PS->getPassEndSpeed() );
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );

      }
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
             SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
     }
     else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
     {
       if( WM->getAgentStamina().getStamina() >     // if stamina high
                            SS->getRecoverDecThr()*SS->getStaminaMax()+800 )
       {
         soc = moveToPos(WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
         ACT->putCommandInQueue( soc );            // move to strategic pos
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else                                        // else watch ball
       {
         ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
     }
     else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     {
       ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     }
     else                                         // nothing to do
       ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
   }
  return soc;
}

/////////////////////////////////////////////////
/////////////////////////////////////////////////

SoccerCommand Player::deMeer5(  )
{

  SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  int           iTmp;

  if( WM->isBeforeKickOff( ) )
  {
    if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    {
      if( WM->isBallKickable() )
      {
        VecPosition posGoal( PITCH_LENGTH / 2.0,
                             ( -1 + 2 * ( WM->getCurrentCycle() % 2 ) ) *
                             0.4 * SS->getGoalWidth() );
        soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal
        Log.log( 100, "take kick off" );
      }
      else
      {
        soc = intercept( false );
        Log.log( 100, "move to ball to take kick-off" );
      }
      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      return soc;
    }
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos( WM->getStrategicPosition() ));
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = searchBall() );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else
  {
     if(posBall.getX()>0){
      setFormationAfterKickOff();
    }else{
      formations->setFormation( FT_433_OFFENSIVE );
    }

    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    }
    else if( WM->isBallKickable())                    // if kickable
    {
      VecPosition posGoal( PITCH_LENGTH/2.0,
              (-1 + 2*(WM->getCurrentCycle()%2)) * 0.4 * SS->getGoalWidth() );
      soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal

      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      Log.log( 100, "kick ball" );
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
             SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
     }
     else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
     {
       if( WM->getAgentStamina().getStamina() >     // if stamina high
                            SS->getRecoverDecThr()*SS->getStaminaMax()+800 )
       {
         soc = moveToPos(WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
         ACT->putCommandInQueue( soc );            // move to strategic pos
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else                                        // else watch ball
       {
         ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
     }
     else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     {
       ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     }
     else                                         // nothing to do
       ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
   }
  return soc;
}

//goalkeeper//
//////////////////////////////////////////
//////////////////////////////////////////
SoccerCommand Player::GK(  )
{
  int i;
  SoccerCommand soc;
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  AngDeg        angBody  = WM->getAgentGlobalBodyAngle();

  // define the top and bottom position of a rectangle in which keeper moves
  static const VecPosition posLeftTop( -PITCH_LENGTH/2.0 +
               0.7*PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/4.0 );
  static const VecPosition posRightTop( -PITCH_LENGTH/2.0 +
               0.7*PENALTY_AREA_LENGTH, +PENALTY_AREA_WIDTH/4.0 );

  // define the borders of this rectangle using the two points.
  static Line  lineFront = Line::makeLineFromTwoPoints(posLeftTop,posRightTop);
  static Line  lineLeft  = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posLeftTop.getY()), posLeftTop );
  static Line  lineRight = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posRightTop.getY()),posRightTop );


  if( WM->isBeforeKickOff( ) )
  {
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos(WM->getStrategicPosition()) );
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = turnBodyToPoint( VecPosition( 0, 0 ), 0 ));
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
    return soc;
  }

  if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
  {                                                // confidence ball too  low
    ACT->putCommandInQueue( searchBall() );        // search ball
    ACT->putCommandInQueue( alignNeckWithBody( ) );
  }
  else if( WM->getPlayMode() == PM_PLAY_ON || WM->isFreeKickThem() ||
           WM->isCornerKickThem() )
  {
    if( WM->isBallCatchable() )
    {
      ACT->putCommandInQueue( soc = catchBall() );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
     else if( WM->isBallKickable() )
    {
       //soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );
       soc = leadingPass(OBJECT_TEAMMATE_2,2.3,DIR_NORTH);
       ACT->putCommandInQueue( soc );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else if( WM->isInOwnPenaltyArea( getInterceptionPointBall( &i, true ) ) &&
             WM->getFastestInSetTo( OBJECT_SET_PLAYERS, OBJECT_BALL, &i ) ==
                                               WM->getAgentObjectType() )
    {
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
    {
      // make line between own goal and the ball
      VecPosition posMyGoal = ( WM->getSide() == SIDE_LEFT )
             ? SoccerTypes::getGlobalPositionFlag( OBJECT_GOAL_L, SIDE_LEFT )
             : SoccerTypes::getGlobalPositionFlag( OBJECT_GOAL_R, SIDE_RIGHT );
      Line lineBall = Line::makeLineFromTwoPoints( WM->getBallPos(), posMyGoal);

      // determine where your front line intersects with the line from ball
      VecPosition posIntersect = lineFront.getIntersection( lineBall );

      // outside rectangle, use line at side to get intersection
      if (posIntersect.isRightOf( posRightTop ) )
        posIntersect = lineRight.getIntersection( lineBall );
      else if (posIntersect.isLeftOf( posLeftTop )  )
        posIntersect = lineLeft.getIntersection( lineBall );

      if( posIntersect.getX() < -49.0 )
        posIntersect.setX( -49.0 );

      // and move to this position
      if( posIntersect.getDistanceTo( WM->getAgentGlobalPosition() ) > 0.5 )
      {
        soc = moveToPos( posIntersect, PS->getPlayerWhenToTurnAngle() );
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else
      {
        ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
  }
  else if( WM->isFreeKickUs() == true || WM->isGoalKickUs() == true )
  {
    if( WM->isBallKickable() )
    {
      if( WM->getTimeSinceLastCatch() == 25 && WM->isFreeKickUs() )
      {
        // move to position with lesser opponents.
        if( WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                                          Circle(posRightTop, 15.0 )) <
            WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                                           Circle(posLeftTop,  15.0 )) )
          soc.makeCommand( CMD_MOVE,posRightTop.getX(),posRightTop.getY(),0.0);
        else
          soc.makeCommand( CMD_MOVE,posLeftTop.getX(), posLeftTop.getY(), 0.0);
        ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() > 28 )
      {
        soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );
        ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() < 25 )
      {
        VecPosition posSide( 0.0, posAgent.getY() );
        if( fabs( (posSide - posAgent).getDirection() - angBody) > 10 )
        {
          soc = turnBodyToPoint( posSide );
          ACT->putCommandInQueue( soc );
        }
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( WM->isGoalKickUs()  )
    {
      ACT->putCommandInQueue( soc = intercept( false ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  else
  {
     ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
     ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  return soc;
}

//////////////////////////////////////////////
//////////////////////////////////////////////

SoccerCommand Player::deMeer5_goalie(  )
{
  int i;
  SoccerCommand soc;
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  AngDeg        angBody  = WM->getAgentGlobalBodyAngle();

  // define the top and bottom position of a rectangle in which keeper moves
  static const VecPosition posLeftTop( -PITCH_LENGTH/2.0 +
               0.7*PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/4.0 );
  static const VecPosition posRightTop( -PITCH_LENGTH/2.0 +
               0.7*PENALTY_AREA_LENGTH, +PENALTY_AREA_WIDTH/4.0 );

  // define the borders of this rectangle using the two points.
  static Line  lineFront = Line::makeLineFromTwoPoints(posLeftTop,posRightTop);
  static Line  lineLeft  = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posLeftTop.getY()), posLeftTop );
  static Line  lineRight = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posRightTop.getY()),posRightTop );


  if( WM->isBeforeKickOff( ) )
  {
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos(WM->getStrategicPosition()) );
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = turnBodyToPoint( VecPosition( 0, 0 ), 0 ));
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
    return soc;
  }

  if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
  {                                                // confidence ball too  low
    ACT->putCommandInQueue( searchBall() );        // search ball
    ACT->putCommandInQueue( alignNeckWithBody( ) );
  }
  else if( WM->getPlayMode() == PM_PLAY_ON || WM->isFreeKickThem() ||
           WM->isCornerKickThem() )
  {
    if( WM->isBallCatchable() )
    {
      ACT->putCommandInQueue( soc = catchBall() );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
     else if( WM->isBallKickable() )
    {
       //soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );
       soc = leadingPass(OBJECT_TEAMMATE_2,2.3,DIR_NORTH);
       ACT->putCommandInQueue( soc );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else if( WM->isInOwnPenaltyArea( getInterceptionPointBall( &i, true ) ) &&
             WM->getFastestInSetTo( OBJECT_SET_PLAYERS, OBJECT_BALL, &i ) ==
                                               WM->getAgentObjectType() )
    {
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
    {
      // make line between own goal and the ball
      VecPosition posMyGoal = ( WM->getSide() == SIDE_LEFT )
             ? SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_L, SIDE_LEFT )
             : SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_R, SIDE_RIGHT);
      Line lineBall = Line::makeLineFromTwoPoints( WM->getBallPos(),posMyGoal);

      // determine where your front line intersects with the line from ball
      VecPosition posIntersect = lineFront.getIntersection( lineBall );

      // outside rectangle, use line at side to get intersection
      if (posIntersect.isRightOf( posRightTop ) )
        posIntersect = lineRight.getIntersection( lineBall );
      else if (posIntersect.isLeftOf( posLeftTop )  )
        posIntersect = lineLeft.getIntersection( lineBall );

      if( posIntersect.getX() < -49.0 )
        posIntersect.setX( -49.0 );

      // and move to this position
      if( posIntersect.getDistanceTo( WM->getAgentGlobalPosition() ) > 0.5 )
      {
        soc = moveToPos( posIntersect, PS->getPlayerWhenToTurnAngle() );
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else
      {
        ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
  }
  else if( WM->isFreeKickUs() == true || WM->isGoalKickUs() == true )
  {
    if( WM->isBallKickable() )
    {
      if( WM->getTimeSinceLastCatch() == 25 && WM->isFreeKickUs() )
      {
        // move to position with lesser opponents.
        if( WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                                          Circle(posRightTop, 15.0 )) <
            WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                                           Circle(posLeftTop,  15.0 )) )
          soc.makeCommand( CMD_MOVE,posRightTop.getX(),posRightTop.getY(),0.0);
        else
          soc.makeCommand( CMD_MOVE,posLeftTop.getX(), posLeftTop.getY(), 0.0);
        ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() > 28 )
      {
        soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );
        ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() < 25 )
      {
        VecPosition posSide( 0.0, posAgent.getY() );
        if( fabs( (posSide - posAgent).getDirection() - angBody) > 10 )
        {
          soc = turnBodyToPoint( posSide );
          ACT->putCommandInQueue( soc );
        }
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( WM->isGoalKickUs()  )
    {
      ACT->putCommandInQueue( soc = intercept( false ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  else
  {
     ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
     ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  return soc;
}


